@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2023 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 package com.artifex.mupdf.fitz;
 
 public class NativeDevice extends Device
@@ -35,7 +57,6 @@ public class NativeDevice extends Device
 	public native final void fillShade(Shade shd, Matrix ctm, float alpha, int cp);
 	public native final void fillImage(Image img, Matrix ctm, float alpha, int cp);
 	public native final void fillImageMask(Image img, Matrix ctm, ColorSpace cs, float[] color, float alpha, int cp);
-	/* FIXME: Why no scissor? */
 	public native final void clipImageMask(Image img, Matrix ctm);
 
 	public native final void popClip();
@@ -48,6 +69,15 @@ public class NativeDevice extends Device
 	public native final int beginTile(Rect area, Rect view, float xstep, float ystep, Matrix ctm, int id);
 	public native final void endTile();
 
+	public native final void renderFlags(int set, int clear);
+	public native final void setDefaultColorSpaces(DefaultColorSpaces dcs);
+
 	public native final void beginLayer(String name);
 	public native final void endLayer();
+
+	public native final void beginStructure(int standard, String raw, int uid);
+	public native final void endStructure();
+
+	public native final void beginMetatext(int meta, String text);
+	public native final void endMetatext();
 }
