@@ -1250,7 +1250,7 @@ static void layout_table(fz_context *ctx, layout_data *ld, fz_html_box *box, fz_
 					if (restart)
 					{
 						restart->end = row;
-						return;
+						goto exit;
 					}
 					else
 					{
@@ -1262,6 +1262,7 @@ static void layout_table(fz_context *ctx, layout_data *ld, fz_html_box *box, fz_
 
 			box->s.layout.b = row->s.layout.b + spacing;
 		}
+		exit:;
 	}
 	fz_always(ctx)
 		fz_free(ctx, colw);
