@@ -1,3 +1,25 @@
+// Copyright (C) 2004-2021 Artifex Software, Inc.
+//
+// This file is part of MuPDF.
+//
+// MuPDF is free software: you can redistribute it and/or modify it under the
+// terms of the GNU Affero General Public License as published by the Free
+// Software Foundation, either version 3 of the License, or (at your option)
+// any later version.
+//
+// MuPDF is distributed in the hope that it will be useful, but WITHOUT ANY
+// WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
+// FOR A PARTICULAR PURPOSE. See the GNU Affero General Public License for more
+// details.
+//
+// You should have received a copy of the GNU Affero General Public License
+// along with MuPDF. If not, see <https://www.gnu.org/licenses/agpl-3.0.en.html>
+//
+// Alternative licensing terms are available from the licensor.
+// For commercial licensing, see <https://www.artifex.com/> or contact
+// Artifex Software, Inc., 1305 Grant Avenue - Suite 200, Novato,
+// CA 94945, U.S.A., +1(415)492-9861, for further information.
+
 #include "mupdf/fitz.h"
 #include "mupdf/pdf.h"
 
@@ -30,6 +52,8 @@ static int ft_font_file_kind(FT_Face face)
 
 static int is_ttc(fz_font *font)
 {
+	if (!font || !font->buffer || font->buffer->len < 4)
+		return 0;
 	return !memcmp(font->buffer->data, "ttcf", 4);
 }
 
@@ -60,6 +84,7 @@ pdf_add_font_file(fz_context *ctx, pdf_document *doc, fz_font *font)
 	fz_buffer *buf = font->buffer;
 	pdf_obj *obj = NULL;
 	pdf_obj *ref = NULL;
+	int drop_buf = 0;
 
 	fz_var(obj);
 	fz_var(ref);
@@ -68,6 +93,13 @@ pdf_add_font_file(fz_context *ctx, pdf_document *doc, fz_font *font)
 	if (font->flags.ft_substitute)
 		return NULL;
 
+	if (is_ttc(font))
+	{
+		buf = NULL;
+		drop_buf = 1;
+		buf = fz_extract_ttf_from_ttc(ctx, font);
+	}
+
 	fz_try(ctx)
 	{
 		size_t len = fz_buffer_storage(ctx, buf, NULL);
@@ -86,7 +118,7 @@ pdf_add_font_file(fz_context *ctx, pdf_document *doc, fz_font *font)
 			if (FT_Get_Sfnt_Table(font->ft_face, FT_SFNT_HEAD))
 				pdf_dict_put(ctx, obj, PDF_NAME(Subtype), PDF_NAME(OpenType));
 			else
-				pdf_dict_put(ctx, obj, PDF_NAME(Subtype), PDF_NAME(Type1C));
+				pdf_dict_put(ctx, obj, PDF_NAME(Subtype), PDF_NAME(CIDFontType0C));
 			break;
 		}
 		ref = pdf_add_object(ctx, doc, obj);
@@ -95,6 +127,8 @@ pdf_add_font_file(fz_context *ctx, pdf_document *doc, fz_font *font)
 	fz_always(ctx)
 	{
 		pdf_drop_obj(ctx, obj);
+		if (drop_buf)
+			fz_drop_buffer(ctx, buf);
 	}
 	fz_catch(ctx)
 	{
@@ -196,7 +230,7 @@ pdf_add_cid_system_info(fz_context *ctx, pdf_document *doc, pdf_obj *fobj, const
 }
 
 /* Different states of starting, same width as last, or consecutive glyph */
-enum { FW_START, FW_SAME, FW_RUN };
+enum { FW_START = 0, FW_SAME, FW_DIFFER };
 
 /* ToDo: Ignore the default sized characters */
 static void
@@ -206,64 +240,72 @@ pdf_add_cid_font_widths(fz_context *ctx, pdf_document *doc, pdf_obj *fobj, fz_fo
 	pdf_obj *run_obj = NULL;
 	pdf_obj *fw;
 	int curr_code;
-	int prev_code;
 	int curr_size;
-	int prev_size;
 	int first_code;
-	int new_first_code;
 	int state = FW_START;
-	int new_state = FW_START;
-	int publish = 0;
 
 	fz_var(run_obj);
 
 	fw = pdf_add_new_array(ctx, doc, 10);
 	fz_try(ctx)
 	{
-		prev_code = 0;
-		prev_size = fz_advance_glyph(ctx, font, 0, 0) * 1000;
-		first_code = prev_code;
+		curr_code = 0;
+		curr_size = fz_advance_glyph(ctx, font, 0, 0) * 1000;
+		first_code = 0;
 
-		for (;;)
+		for (curr_code = 1; curr_code < face->num_glyphs; curr_code++)
 		{
-			curr_code = prev_code + 1;
-			if (curr_code >= face->num_glyphs)
-				break;
+			int prev_size = curr_size;
+
 			curr_size = fz_advance_glyph(ctx, font, curr_code, 0) * 1000;
 
+			/* So each time around the loop when we reach here, we have sizes
+			 * for curr_code-1 (prev_size) and curr_code (curr_size), neither
+			 * of which have been published yet. By the time we reach the end
+			 * of the loop we must have disposed of prev_size. */
 			switch (state)
 			{
 			case FW_SAME:
+				/* Until now, we've been in a run of identical values, extending
+				 * from first_code to curr_code-1. If the current and prev sizes
+				 * match, then this now extends from first_code to curr_code and
+				 * we don't need to do anything. If not, we need to flush and
+				 * restart. */
 				if (curr_size != prev_size)
 				{
-					/* End of same widths for consecutive ids. Current will
-					 * be pushed as prev. below during next iteration */
-					publish = 1;
-					if (curr_code < face->num_glyphs)
-						run_obj = pdf_new_array(ctx, doc, 10);
-					new_state = FW_RUN;
-					/* And the new first code is our current code */
-					new_first_code = curr_code;
+					/* Add three entries. First cid, last cid and width */
+					pdf_array_push_int(ctx, fw, first_code);
+					pdf_array_push_int(ctx, fw, curr_code-1);
+					pdf_array_push_int(ctx, fw, prev_size);
+					/* And the new first code is our current code. */
+					first_code = curr_code;
+					state = FW_START;
 				}
 				break;
-			case FW_RUN:
+			case FW_DIFFER:
+				/* Until now, we've been in a run of differing values, extending
+				 * from first_code to curr_code-1 (though prev_size, the size for
+				 * curr_code-1 has not yet been pushed). */
 				if (curr_size == prev_size)
 				{
-					/* Same width, so start a new same entry starting with
-					 * the previous code. i.e. the prev size is not put
-					 * in the run */
-					publish = 1;
-					new_state = FW_SAME;
-					new_first_code = prev_code;
+					/* Same width, so flush the run of differences. */
+					pdf_array_push_int(ctx, fw, first_code);
+					pdf_array_push(ctx, fw, run_obj);
+					pdf_drop_obj(ctx, run_obj);
+					run_obj = NULL;
+					/* Start a new 'same' entry starting with curr_code-1.
+					 * i.e. the prev size is not put in the run. */
+					state = FW_SAME;
+					first_code = curr_code-1;
 				}
 				else
 				{
-					/* Add prev size to run_obj */
+					/* Continue our differing run by adding prev size to run_obj. */
 					pdf_array_push_int(ctx, run_obj, prev_size);
 				}
 				break;
 			case FW_START:
-				/* Starting fresh. Determine our state */
+				/* Starting fresh. Determine our state. */
 				if (curr_size == prev_size)
 				{
 					state = FW_SAME;
@@ -272,55 +314,38 @@ pdf_add_cid_font_widths(fz_context *ctx, pdf_document *doc, pdf_obj *fobj, fz_fo
 				{
 					run_obj = pdf_new_array(ctx, doc, 10);
 					pdf_array_push_int(ctx, run_obj, prev_size);
-					state = FW_RUN;
+					state = FW_DIFFER;
 				}
-				new_first_code = prev_code;
 				break;
 			}
-
-			if (publish || curr_code == face->num_glyphs)
-			{
-				switch (state)
-				{
-				case FW_SAME:
-					/* Add three entries. First cid, last cid and width */
-					pdf_array_push_int(ctx, fw, first_code);
-					pdf_array_push_int(ctx, fw, prev_code);
-					pdf_array_push_int(ctx, fw, prev_size);
-					break;
-				case FW_RUN:
-					if (pdf_array_len(ctx, run_obj) > 0)
-					{
-						pdf_array_push_int(ctx, fw, first_code);
-						pdf_array_push(ctx, fw, run_obj);
-					}
-					pdf_drop_obj(ctx, run_obj);
-					run_obj = NULL;
-					break;
-				case FW_START:
-					/* Lone wolf. Not part of a consecutive run */
-					pdf_array_push_int(ctx, fw, prev_code);
-					pdf_array_push_int(ctx, fw, prev_code);
-					pdf_array_push_int(ctx, fw, prev_size);
-					break;
-				}
-
-				if (curr_code < face->num_glyphs)
-				{
-					state = new_state;
-					first_code = new_first_code;
-					publish = 0;
-				}
-			}
-
-			prev_size = curr_size;
-			prev_code = curr_code;
 		}
 
-		if (pdf_array_len(ctx, run_obj) > 0)
+		/* So curr_code-1 is the last valid char, and curr_size was its size. */
+		switch (state)
 		{
+		case FW_SAME:
+			/* We have an unflushed run of same entries. */
+			if (first_code != curr_code-1)
+			{
+				pdf_array_push_int(ctx, fw, first_code);
+				pdf_array_push_int(ctx, fw, curr_code-1);
+				pdf_array_push_int(ctx, fw, curr_size);
+			}
+			break;
+		case FW_DIFFER:
+			/* We have not yet pushed curr_size to the object. */
 			pdf_array_push_int(ctx, fw, first_code);
+			pdf_array_push_int(ctx, run_obj, curr_size);
 			pdf_array_push(ctx, fw, run_obj);
+			pdf_drop_obj(ctx, run_obj);
+			run_obj = NULL;
+			break;
+		case FW_START:
+			/* Lone wolf! */
+			pdf_array_push_int(ctx, fw, curr_code-1);
+			pdf_array_push_int(ctx, fw, curr_code-1);
+			pdf_array_push_int(ctx, fw, curr_size);
+			break;
 		}
 
 		if (font->width_table != NULL)
@@ -547,9 +572,9 @@ pdf_add_cid_font(fz_context *ctx, pdf_document *doc, fz_font *font)
 	pdf_obj *fobj = NULL;
 	pdf_obj *fref = NULL;
 	pdf_obj *dfonts = NULL;
-	unsigned char digest[16];
+	pdf_font_resource_key key;
 
-	fref = pdf_find_font_resource(ctx, doc, PDF_CID_FONT_RESOURCE, 0, font, digest);
+	fref = pdf_find_font_resource(ctx, doc, PDF_CID_FONT_RESOURCE, 0, font, &key);
 	if (fref)
 		return fref;
 
@@ -565,7 +590,7 @@ pdf_add_cid_font(fz_context *ctx, pdf_document *doc, fz_font *font)
 		dfonts = pdf_dict_put_array(ctx, fobj, PDF_NAME(DescendantFonts), 1);
 		pdf_array_push_drop(ctx, dfonts, pdf_add_descendant_cid_font(ctx, doc, font));
 
-		fref = pdf_insert_font_resource(ctx, doc, digest, fobj);
+		fref = pdf_insert_font_resource(ctx, doc, &key, fobj);
 	}
 	fz_always(ctx)
 		pdf_drop_obj(ctx, fobj);
@@ -624,9 +649,9 @@ pdf_add_simple_font(fz_context *ctx, pdf_document *doc, fz_font *font, int encod
 	pdf_obj *fobj = NULL;
 	pdf_obj *fref = NULL;
 	const char **enc;
-	unsigned char digest[16];
+	pdf_font_resource_key key;
 
-	fref = pdf_find_font_resource(ctx, doc, PDF_SIMPLE_FONT_RESOURCE, encoding, font, digest);
+	fref = pdf_find_font_resource(ctx, doc, PDF_SIMPLE_FONT_RESOURCE, encoding, font, &key);
 	if (fref)
 		return fref;
 
@@ -665,7 +690,7 @@ pdf_add_simple_font(fz_context *ctx, pdf_document *doc, fz_font *font, int encod
 				pdf_add_simple_font_widths(ctx, doc, fobj, font, enc);
 		}
 
-		fref = pdf_insert_font_resource(ctx, doc, digest, fobj);
+		fref = pdf_insert_font_resource(ctx, doc, &key, fobj);
 	}
 	fz_always(ctx)
 	{
@@ -679,10 +704,10 @@ pdf_add_simple_font(fz_context *ctx, pdf_document *doc, fz_font *font, int encod
 int
 pdf_font_writing_supported(fz_font *font)
 {
-	if (font->ft_face == NULL || font->buffer == NULL || font->buffer->len < 4)
+	if (font->ft_face == NULL || font->buffer == NULL || font->buffer->len < 4 || !font->flags.embed || font->flags.never_embed)
 		return 0;
 	if (is_ttc(font))
-		return 0;
+		return 1;
 	if (is_truetype(font->ft_face))
 		return 1;
 	if (is_postscript(font->ft_face))
@@ -696,7 +721,7 @@ pdf_add_cjk_font(fz_context *ctx, pdf_document *doc, fz_font *fzfont, int script
 	pdf_obj *fref, *font, *subfont, *fontdesc;
 	pdf_obj *dfonts;
 	fz_rect bbox = { -200, -200, 1200, 1200 };
-	unsigned char digest[16];
+	pdf_font_resource_key key;
 	int flags;
 
 	const char *basefont, *encoding, *ordering;
@@ -737,7 +762,7 @@ pdf_add_cjk_font(fz_context *ctx, pdf_document *doc, fz_font *fzfont, int script
 	if (serif)
 		flags |= PDF_FD_SERIF;
 
-	fref = pdf_find_font_resource(ctx, doc, PDF_CJK_FONT_RESOURCE, script, fzfont, digest);
+	fref = pdf_find_font_resource(ctx, doc, PDF_CJK_FONT_RESOURCE, script, fzfont, &key);
 	if (fref)
 		return fref;
 
@@ -769,7 +794,7 @@ pdf_add_cjk_font(fz_context *ctx, pdf_document *doc, fz_font *fzfont, int script
 			}
 		}
 
-		fref = pdf_insert_font_resource(ctx, doc, digest, font);
+		fref = pdf_insert_font_resource(ctx, doc, &key, font);
 	}
 	fz_always(ctx)
 		pdf_drop_obj(ctx, font);
@@ -778,3 +803,10 @@ pdf_add_cjk_font(fz_context *ctx, pdf_document *doc, fz_font *fzfont, int script
 
 	return fref;
 }
+
+pdf_obj *
+pdf_add_substitute_font(fz_context *ctx, pdf_document *doc, fz_font *font)
+{
+	fz_throw(ctx, FZ_ERROR_GENERIC, "substitute font creation is not implemented yet");
+	return NULL;
+}
